=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2008-05-16 05:59:21 +0000
+++ bzrlib/errors.py	2008-05-21 19:02:34 +0000
@@ -2253,6 +2253,17 @@
             " Please set BZR_SSH environment variable.")
 
 
+class GhostRevisionsHaveNoRevno(BzrError):
+    """When searching for revnos, if we encounter a ghost, we are stuck"""
+
+    _fmt = ("Could not determine revno for {%(revision_id)s} because"
+            " its ancestry shows a ghost at {%(ghost_revision_id)s}")
+
+    def __init__(self, revision_id, ghost_revision_id):
+        self.revision_id = revision_id
+        self.ghost_revision_id = ghost_revision_id
+
+        
 class GhostRevisionUnusableHere(BzrError):
 
     _fmt = "Ghost revision {%(revision_id)s} cannot be used here."

=== modified file 'bzrlib/graph.py'
--- bzrlib/graph.py	2008-05-21 05:46:23 +0000
+++ bzrlib/graph.py	2008-05-21 19:02:34 +0000
@@ -212,6 +212,32 @@
         right = searchers[1].seen
         return (left.difference(right), right.difference(left))
 
+    def find_revno(self, target_revision_id, known_revision_ids):
+        """Determine the revno for target_revision_id.
+
+        :param target_revision_id: A revision_id which we would like to know
+            the revno for.
+        :param known_revision_ids: [(revision_id, revno)] A list of known
+            revno, revision_id tuples. We'll use this to seed the search.
+        """
+        # Map from revision_ids to a known value for their revno
+        known_revnos = dict(known_revision_ids)
+        cur_tip = target_revision_id
+        num_steps = 0
+        NULL_REVISION = revision.NULL_REVISION
+
+        while cur_tip != NULL_REVISION:
+            parent_map = self.get_parent_map([cur_tip])
+            parents = parent_map.get(cur_tip, None)
+            if not parents: # An empty list, or None is still a ghost
+                raise errors.GhostRevisionsHaveNoRevno(target_revision_id,
+                                                       cur_tip)
+            num_steps += 1
+            cur_tip = parents[0]
+
+        # We reached NULL_REVISION, so the total distance is num_steps + 1
+        return num_steps
+
     def find_unique_ancestors(self, unique_revision, common_revisions):
         """Find the unique ancestors for a revision versus others.
 

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2008-05-11 23:49:50 +0000
+++ bzrlib/tests/test_errors.py	2008-05-21 19:02:34 +0000
@@ -52,6 +52,12 @@
         self.assertEqualDiff('The prefix foo is in the help search path twice.',
             str(error))
 
+    def test_ghost_revisions_have_no_revno(self):
+        error = errors.GhostRevisionsHaveNoRevno('target', 'ghost_rev')
+        self.assertEqualDiff("Could not determine revno for {target} because"
+                             " its ancestry shows a ghost at {ghost_rev}",
+                             str(error))
+
     def test_incompatibleAPI(self):
         error = errors.IncompatibleAPI("module", (1, 2, 3), (4, 5, 6), (7, 8, 9))
         self.assertEqualDiff(

=== modified file 'bzrlib/tests/test_graph.py'
--- bzrlib/tests/test_graph.py	2008-05-06 20:54:31 +0000
+++ bzrlib/tests/test_graph.py	2008-05-21 19:02:34 +0000
@@ -1228,6 +1228,40 @@
             ['h', 'i', 'j', 'y'], 'j', ['z'])
 
 
+class TestGraphFindRevno(tests.TestCase):
+    """Test an api that should be able to compute a revno"""
+
+    def make_graph(self, ancestors):
+        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))
+
+    def assertFindRevno(self, revno, graph, target_id, known_ids):
+        """Assert the output of Graph.find_revno()"""
+        actual = graph.find_revno(target_id, known_ids)
+        self.assertEqual(revno, actual)
+
+    def test_nothing_known(self):
+        graph = self.make_graph(ancestry_1)
+        self.assertFindRevno(1, graph, 'rev1', [])
+        self.assertFindRevno(2, graph, 'rev2a', [])
+        self.assertFindRevno(2, graph, 'rev2b', [])
+        self.assertFindRevno(3, graph, 'rev3', [])
+        self.assertFindRevno(4, graph, 'rev4', [])
+
+    def test_rev_is_ghost(self):
+        graph = self.make_graph(ancestry_1)
+        e = self.assertRaises(errors.GhostRevisionsHaveNoRevno,
+                              graph.find_revno, 'rev_missing', [])
+        self.assertEqual('rev_missing', e.revision_id)
+        self.assertEqual('rev_missing', e.ghost_revision_id)
+
+    def test_ancestor_is_ghost(self):
+        graph = self.make_graph({'rev':['parent']})
+        e = self.assertRaises(errors.GhostRevisionsHaveNoRevno,
+                              graph.find_revno, 'rev', [])
+        self.assertEqual('rev', e.revision_id)
+        self.assertEqual('parent', e.ghost_revision_id)
+
+
 class TestCachingParentsProvider(tests.TestCase):
 
     def setUp(self):

